"""The Python interpreter may switch between threads inbetween bytecode
execution.  Bytecode execution in fastcache may occur during:
(1) Calls to make_key which will call the __hash__ methods of the args and
(2) `PyDict_Get(Set)Item` calls rely on Python comparisons (i.e, __eq__)
    to determine if a match has been found

A good test for threadsafety is then to cache a function which takes user
defined Python objects that have __hash__ and __eq__ methods which live in
Python land rather built-in land.

The test should not only ensure that the correct result is acheived (and no
segfaults) but also assess memory leaks.

The thread switching interval can be altered using sys.setswitchinterval.
"""

class PythonInt:
    """ Wrapper for an integer with python versions of __eq__ and __hash__."""

    def __init__(self, val):
        self.value = val

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        # only compare with other instances of PythonInt
        if not isinstance(other, PythonInt):
            raise TypeError("PythonInt cannot be compared to %s" % type(other))
        return self.value == other.value

from random import randint
import unittest
from fastcache import clru_cache as lru_cache
from threading import Thread
try:
    from sys import setswitchinterval as setinterval
except ImportError:
    from sys import setcheckinterval
    def setinterval(i):
        return setcheckinterval(int(i))


def run_threads(threads):
    for t in threads:
        t.start()
    for t in threads:
        t.join()

CACHE_SIZE=301
FIB=CACHE_SIZE-1
RAND_MIN, RAND_MAX = 1, 10

@lru_cache(maxsize=CACHE_SIZE, typed=False)
def fib(n):
    """Terrible Fibonacci number generator."""
    v = n.value
    return v if v < 2 else fib(PythonInt(v-1)) + fib(PythonInt(v-2))

# establish correct result from single threaded exectution
RESULT = fib(PythonInt(FIB))

def run_fib_with_clear(r):
    """ Run Fibonacci generator r times. """
    for i in range(r):
        if randint(RAND_MIN, RAND_MAX) == RAND_MIN:
            fib.cache_clear()
        res = fib(PythonInt(FIB))
        if RESULT != res:
            raise ValueError("Expected %d, Got %d" % (RESULT, res))

def run_fib_with_stats(r):
    """ Run Fibonacci generator r times. """
    for i in range(r):
        res = fib(PythonInt(FIB))
        if RESULT != res:
            raise ValueError("Expected %d, Got %d" % (RESULT, res))


class Test_Threading(unittest.TestCase):
    """ Threadsafety Tests for lru_cache. """

    def setUp(self):
        setinterval(1e-6)
        self.numthreads = 4
        self.repeat = 1000

    def test_thread_random_cache_clears(self):
        """ randomly clear the cache during calls to fib. """

        threads = [Thread(target=run_fib_with_clear, args=(self.repeat, ))
                   for _ in range(self.numthreads)]
        run_threads(threads)
        # if we have gotten this far no exceptions have been raised
        self.assertEqual(0, 0)

    def test_thread_cache_info(self):
        """ Run thread safety test to make sure the cache statistics
        are correct."""
        fib.cache_clear()
        threads = [Thread(target=run_fib_with_stats, args=(self.repeat, ))
                   for _ in range(self.numthreads)]
        run_threads(threads)

        hits, misses, maxsize, currsize = fib.cache_info()
        self.assertEqual(misses, CACHE_SIZE)
        self.assertEqual(currsize, CACHE_SIZE)



# Copied from python src Python-3.4.0/Lib/test/test_functools.py

import abc
import collections
from itertools import permutations
import pickle
from random import choice
import sys
import unittest
import fastcache
import functools

try:
    from functools import _CacheInfo
except ImportError:
    _CacheInfo = collections.namedtuple("CacheInfo", 
                    ["hits", "misses", "maxsize", "currsize"])

class TestLRU(unittest.TestCase):

    def test_lru(self):
        def orig(x, y):
            return 3 * x + y
        f = fastcache.clru_cache(maxsize=20)(orig)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(maxsize, 20)
        self.assertEqual(currsize, 0)
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 0)

        domain = range(5)
        for i in range(1000):
            x, y = choice(domain), choice(domain)
            actual = f(x, y)
            expected = orig(x, y)
            self.assertEqual(actual, expected)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertTrue(hits > misses)
        self.assertEqual(hits + misses, 1000)
        self.assertEqual(currsize, 20)

        f.cache_clear()   # test clearing
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 0)
        self.assertEqual(currsize, 0)
        f(x, y)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 1)
        self.assertEqual(currsize, 1)

        # Test bypassing the cache
        if hasattr(self, 'assertIs'):
            self.assertIs(f.__wrapped__, orig)
        f.__wrapped__(x, y)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 1)
        self.assertEqual(currsize, 1)

        # test size zero (which means "never-cache")
        @fastcache.clru_cache(0)
        def f():
            #nonlocal f_cnt
            f_cnt[0] += 1
            return 20
        self.assertEqual(f.cache_info().maxsize, 0)
        f_cnt = [0]
        for i in range(5):
            self.assertEqual(f(), 20)
        self.assertEqual(f_cnt, [5])
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 5)
        self.assertEqual(currsize, 0)

        # test size one
        @fastcache.clru_cache(1)
        def f():
            #nonlocal f_cnt
            f_cnt[0] += 1
            return 20
        self.assertEqual(f.cache_info().maxsize, 1)
        f_cnt[0] = 0
        for i in range(5):
            self.assertEqual(f(), 20)
        self.assertEqual(f_cnt, [1])
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 4)
        self.assertEqual(misses, 1)
        self.assertEqual(currsize, 1)

        # test size two
        @fastcache.clru_cache(2)
        def f(x):
            #nonlocal f_cnt
            f_cnt[0] += 1
            return x*10
        self.assertEqual(f.cache_info().maxsize, 2)
        f_cnt[0] = 0
        for x in 7, 9, 7, 9, 7, 9, 8, 8, 8, 9, 9, 9, 8, 8, 8, 7:
            #    *  *              *                          *
            self.assertEqual(f(x), x*10)
        self.assertEqual(f_cnt, [4])
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 12)
        self.assertEqual(misses, 4)
        self.assertEqual(currsize, 2)

    def test_lru_with_maxsize_none(self):
        @fastcache.clru_cache(maxsize=None)
        def fib(n):
            if n < 2:
                return n
            return fib(n-1) + fib(n-2)
        self.assertEqual([fib(n) for n in range(16)],
            [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610])
        self.assertEqual(fib.cache_info(),
            _CacheInfo(hits=28, misses=16, maxsize=None, currsize=16))
        fib.cache_clear()
        self.assertEqual(fib.cache_info(),
            _CacheInfo(hits=0, misses=0, maxsize=None, currsize=0))

    def test_lru_with_exceptions(self):
        # Verify that user_function exceptions get passed through without
        # creating a hard-to-read chained exception.
        # http://bugs.python.org/issue13177
        for maxsize in (None, 128):
            @fastcache.clru_cache(maxsize)
            def func(i):
                return 'abc'[i]
            self.assertEqual(func(0), 'a')
            try:
                with self.assertRaises(IndexError) as cm:
                    func(15)
                # Does not have this attribute in Py2
                if hasattr(cm.exception,'__context__'):
                    self.assertIsNone(cm.exception.__context__)
                # Verify that the previous exception did not result in a cached entry
                with self.assertRaises(IndexError):
                    func(15)
            except TypeError:
                # py26 unittest wants assertRaises called with another arg
                if sys.version_info[:2] != (2, 6):
                    raise
                else:
                    pass

    def test_lru_with_types(self):
        for maxsize in (None, 128):
            @fastcache.clru_cache(maxsize=maxsize, typed=True)
            def square(x):
                return x * x
            self.assertEqual(square(3), 9)
            self.assertEqual(type(square(3)), type(9))
            self.assertEqual(square(3.0), 9.0)
            self.assertEqual(type(square(3.0)), type(9.0))
            self.assertEqual(square(x=3), 9)
            self.assertEqual(type(square(x=3)), type(9))
            self.assertEqual(square(x=3.0), 9.0)
            self.assertEqual(type(square(x=3.0)), type(9.0))
            self.assertEqual(square.cache_info().hits, 4)
            self.assertEqual(square.cache_info().misses, 4)

    def test_lru_with_keyword_args(self):
        @fastcache.clru_cache()
        def fib(n):
            if n < 2:
                return n
            return fib(n=n-1) + fib(n=n-2)
        self.assertEqual(
            [fib(n=number) for number in range(16)],
            [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610]
        )
        self.assertEqual(fib.cache_info(),
            _CacheInfo(hits=28, misses=16, maxsize=128, currsize=16))
        fib.cache_clear()
        self.assertEqual(fib.cache_info(),
            _CacheInfo(hits=0, misses=0, maxsize=128, currsize=0))

    def test_lru_with_keyword_args_maxsize_none(self):
        @fastcache.clru_cache(maxsize=None)
        def fib(n):
            if n < 2:
                return n
            return fib(n=n-1) + fib(n=n-2)
        self.assertEqual([fib(n=number) for number in range(16)],
            [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610])
        self.assertEqual(fib.cache_info(),
            _CacheInfo(hits=28, misses=16, maxsize=None, currsize=16))
        fib.cache_clear()
        self.assertEqual(fib.cache_info(),
            _CacheInfo(hits=0, misses=0, maxsize=None, currsize=0))

    def test_need_for_rlock(self):
        # This will deadlock on an LRU cache that uses a regular lock

        @fastcache.clru_cache(maxsize=10)
        def test_func(x):
            'Used to demonstrate a reentrant lru_cache call within a single thread'
            return x

        class DoubleEq:
            'Demonstrate a reentrant lru_cache call within a single thread'
            def __init__(self, x):
                self.x = x
            def __hash__(self):
                return self.x
            def __eq__(self, other):
                if self.x == 2:
                    test_func(DoubleEq(1))
                return self.x == other.x

        test_func(DoubleEq(1))                      # Load the cache
        test_func(DoubleEq(2))                      # Load the cache
        self.assertEqual(test_func(DoubleEq(2)),    # Trigger a re-entrant __eq__ call
                         DoubleEq(2))               # Verify the correct return value



import pytest
import fastcache
import itertools
import warnings

try:
    itertools.count(start=0, step=-1)
    count = itertools.count
except TypeError:
    def count(start=0, step=1):
        i = step-1
        for j, c in enumerate(itertools.count(start)):
            yield c + i*j

def arg_gen(min=1, max=100, repeat=3):
    for i in range(min, max):
        for r in range(repeat):
            for j, k in zip(range(i), count(i, -1)):
                yield j, k

@pytest.fixture(scope='module', params=[fastcache.clru_cache,
                                        fastcache.lru_cache])
def cache(request):
    param = request.param
    return param


def test_function_attributes(cache):
    """ Simple tests for attribute preservation. """

    def tfunc(a, b):
        """test function docstring."""
        return a + b
    cfunc = cache()(tfunc)
    assert cfunc.__doc__ == tfunc.__doc__
    assert hasattr(cfunc, 'cache_info')
    assert hasattr(cfunc, 'cache_clear')
    assert hasattr(cfunc, '__wrapped__')


def test_function_cache(cache):
    """ Test that cache returns appropriate values. """

    cat_tuples = [True]

    def tfunc(a, b, c=None):
        if (cat_tuples[0] == True):
            return (a, b, c) + (c, a)
        else:
            return 2*a-10*b

    cfunc = cache(maxsize=100, state=cat_tuples)(tfunc)

    for i, j in arg_gen(max=75, repeat=5):
        assert cfunc(i, j) == tfunc(i, j)

    # change extra state
    cat_tuples[0] = False

    for i, j in arg_gen(max=75, repeat=5):
        assert cfunc(i, j) == tfunc(i, j)

    # test dict state
    d = {}
    cfunc = cache(maxsize=100, state=d)(tfunc)
    cfunc(1, 2)
    assert cfunc.cache_info().misses == 1
    d['a'] = 42
    cfunc(1, 2)
    assert cfunc.cache_info().misses == 2
    cfunc(1, 2)
    assert cfunc.cache_info().misses == 2
    assert cfunc.cache_info().hits == 1
    d.clear()
    cfunc(1, 2)
    assert cfunc.cache_info().misses == 2
    assert cfunc.cache_info().hits == 2
    d['a'] = 44
    cfunc(1, 2)
    assert cfunc.cache_info().misses == 3

def test_memory_leaks(cache):
    """ Longer running test to check for memory leaks. """

    def tfunc(a, b, c):
        return (a-1, 2*c) + (10*b-1, a*b, a*b+c)

    cfunc = cache(maxsize=2000)(tfunc)

    for i, j in arg_gen(max=1500, repeat=5):
        assert cfunc(i, j, c=i-j) == tfunc(i, j, c=i-j)

def test_warn_unhashable_args(cache, recwarn):
    """ Function arguments must be hashable. """

    @cache(unhashable='warning')
    def f(a, b):
        return (a, ) + (b, )

    with warnings.catch_warnings() :
        warnings.simplefilter("always")
        assert f([1], 2) == f.__wrapped__([1], 2)
        w = recwarn.pop(UserWarning)
        assert issubclass(w.category, UserWarning)
        assert "Unhashable arguments cannot be cached" in str(w.message)
        assert w.filename
        assert w.lineno


def test_ignore_unhashable_args(cache):
    """ Function arguments must be hashable. """

    @cache(unhashable='ignore')
    def f(a, b):
        return (a, ) + (b, )

    assert f([1], 2) == f.__wrapped__([1], 2)

def test_default_unhashable_args(cache):
    @cache()
    def f(a, b):
        return (a, ) + (b, )

    with pytest.raises(TypeError):
        f([1], 2)

    @cache(unhashable='error')
    def f(a, b):
        pass
    with pytest.raises(TypeError):
        f([1], 2)

def test_state_type(cache):
    """ State must be a list or dict. """
    f = lambda x : x
    with pytest.raises(TypeError):
        cache(state=(1, ))(f)
    with pytest.raises(TypeError):
        cache(state=-1)(f)

def test_typed_False(cache):
    """ Verify typed==False. """

    @cache(typed=False)
    def cfunc(a, b):
        return a+b

    # initialize cache with integer args
    cfunc(1, 2)
    assert cfunc(1, 2) is cfunc(1.0, 2)
    assert cfunc(1, 2) is cfunc(1, 2.0)
    # test keywords
    cfunc(1, b=2)
    assert cfunc(1,b=2) is cfunc(1.0,b=2)
    assert cfunc(1,b=2) is cfunc(1,b=2.0)

def test_typed_True(cache):
    """ Verify typed==True. """

    @cache(typed=True)
    def cfunc(a, b):
        return a+b

    assert cfunc(1, 2) is not cfunc(1.0, 2)
    assert cfunc(1, 2) is not cfunc(1, 2.0)
    # test keywords
    assert cfunc(1,b=2) is not cfunc(1.0,b=2)
    assert cfunc(1,b=2) is not cfunc(1,b=2.0)

def test_dynamic_attribute(cache):
    f = lambda x : x
    cfunc = cache()(f)
    cfunc.new_attr = 5
    assert cfunc.new_attr == 5
